/*
 * Copyright (c) 2003-2007 Erez Zadok
 * Copyright (c) 2003-2007 Stony Brook University
 * Copyright (c) 2003-2007 The Research Foundation of State University of New York
 * Copyright (c) 2004-2007 Nikolai Joukov
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>

/* set to 0 to have no timeline and just the totals */
#define MAX_DISTR_TIME 600
#define MAX_JITTER 200
#define RESOLUTION 31

struct mjitter_pd {
	unsigned long long init_cycle;
	unsigned long long read_cycle;
	struct proc_dir_entry *mjitter_proc_root;
	unsigned long distribution[MAX_DISTR_TIME + 1][MAX_JITTER + 1];
};

/* For now just use a static variable. */
static struct mjitter_pd* this = NULL;

#define INIT_CYCLE (context->init_cycle)

#ifdef __ia64__
#define RDTSC(qp) \
	do { \
		unsigned long result; \
		__asm__ __volatile__("mov %0=ar.itc" : "=r" (result) :: "memory"); \
		qp = result; \
	} while (0);

#else

#define RDTSC(qp) \
	do { \
		unsigned long lowPart, highPart; \
		__asm__ __volatile__("rdtsc" : "=a" (lowPart), "=d" (highPart)); \
		qp = (((unsigned long long) highPart) << 32) | lowPart; \
	} while (0);

#endif

static void mjitter_stat_reset(void)
{
	memset(this, 0, sizeof(struct mjitter_pd));
	RDTSC(this->init_cycle);
}

#define PROC_NAME "fs_replayfs_jitter"

#define ADD_TO_BUFFER(str) \
	do { \
		char* tmp; \
		len = strlen((str)); \
		if (*ppos < total + len) { \
			if (count < done + len) { \
				len = count - done; \
			} \
			if (total < *ppos) { \
				len -= *ppos - total;\
				tmp = (str) + *ppos - total; \
			} else {\
				tmp = (str); \
			} \
			if (copy_to_user(buf, tmp, len)) { \
				done = -EFAULT; \
				goto out; \
			} \
			buf+= len; \
			done+= len; \
			if (count == done) \
				goto out; \
		} \
		total+= len; \
	} while(0);


static ssize_t mjitter_proc_read(struct file *file, char *buf, size_t count, loff_t *ppos) {
	char localbuf[1024];
	int len, done = 0, total = 0;

	unsigned int i, ii, iii, last_tick;

	if (!this->read_cycle)
		RDTSC(this->read_cycle);

	last_tick = (unsigned int)((this->read_cycle - this->init_cycle) >> RESOLUTION);
	if (last_tick > MAX_DISTR_TIME)
		last_tick = MAX_DISTR_TIME;

	for (ii = 0; ii < last_tick + 1; ii++) {
		len = 0;
		for (iii = 0; iii < MAX_JITTER + 1; iii++) {
			if (this->distribution[ii][iii] != 0)
				len++;
		}
		if (len) {
			for (iii = 0; iii < MAX_JITTER + 1; iii++) {
				sprintf(localbuf, " %lu", this->distribution[ii][iii]);
				ADD_TO_BUFFER(localbuf);
			}
			ADD_TO_BUFFER("\n");
		} else {
			ADD_TO_BUFFER("-\n");
		}
	}
	this->read_cycle = 0;
out:
	if (done > 0)
		*ppos += done;
	return done;
}

static ssize_t mjitter_proc_write(struct file *file, const char *buf, size_t count, loff_t *ppos) {
	mjitter_stat_reset();
	return count;
}

struct file_operations proc_mjitter_file_operations = {
	read: mjitter_proc_read,
	write: mjitter_proc_write
};

static int mjitter_proc_create(void)
{
	struct proc_dir_entry *proc_de;
	int err = 0;

	proc_de = create_proc_entry(PROC_NAME, 0, NULL);
	if (!proc_de) {
		printk(KERN_ERR "Adding proc entry failed\n");
		goto out;
	}
	proc_de->owner = THIS_MODULE;
	proc_de->data = (void *)this;
	proc_de->proc_fops = &proc_mjitter_file_operations;
out:
	return err;
}

static void mjitter_proc_destroy(void)
{
	remove_proc_entry(PROC_NAME, NULL);
}


int init_module()
{
	int err = 0;

	if (this)
		goto out;

	err = -ENOMEM;

	if (sizeof(struct mjitter_pd) > PAGE_SIZE)
		this = vmalloc(sizeof(struct mjitter_pd));
	else
		this = kmalloc(sizeof(struct mjitter_pd), GFP_KERNEL);

	if (!this)
		goto out;

	mjitter_stat_reset();

	err = mjitter_proc_create();

out:
	return err;
}

void cleanup_module(void)
{
	mjitter_proc_destroy();

	if (this) {
		if (sizeof(struct mjitter_pd) > PAGE_SIZE)
			vfree(this);
		else
			kfree(this);
		this = NULL;
	}
}

/* we use RDTSC to make sure our timing == ext2a timing */
void mjitter_add_value(unsigned int error)
{
	unsigned int ii;
	unsigned long long l;
	RDTSC(l);

	if (error > MAX_JITTER)
		error = MAX_JITTER;

	ii = (int)((l - this->init_cycle) >> RESOLUTION);
	if (ii > MAX_DISTR_TIME)
		ii = MAX_DISTR_TIME;
	this->distribution[ii][error]++;
}
EXPORT_SYMBOL(mjitter_add_value);

//EXPORT_NO_SYMBOLS;
MODULE_AUTHOR("Nikolai Joukov & Timothy T. Wong & Erez Zadok");
MODULE_DESCRIPTION("TraceFS replay engine");
MODULE_LICENSE("GPL");


