/*
 * Copyright (c) 2004-2007 Erez Zadok
 * Copyright (c) 2004-2007 Stony Brook University
 * Copyright (c) 2004-2007 The Research Foundation of State University of New York
 * Copyright (c) 2004-2007 Nikolai Joukov
 * Copyright (c) 2004-2005 Akshat Aranya
 * Copyright (c) 2004-2005 Timothy Wong
 * Copyright (c) 2006-2007 Gopala Surya
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>
#include "replay.h"

#include "tracereader.h"
#include "namehash.h"
#include "pidhash.h"
#include "od_cksum.h"
#include <unistd.h>
#include "util.h"
#include "trfs_pre.h"
#include "hash.h"
#include "thread_manager.h"

#include <linux/limits.h>
#include <linux/stat.h>
#include <linux/errno.h>

static int indent_level = 0;
//pre-compiler logs all commands it processed into /tmp/prefile.
//This is for debugging and needs to be cleaned up once it gets matured.
FILE *prefp ;

#define PID_INVALID (-1)
#define NAMECOMP_NUMFIELDS 3
struct namecomp {
	 int parent_inum;
	 int child_inum;
	 char *name;
};

#define SYSCALL_NUMFIELDS 2
#define SYSCALL_NAME_LENGTH 64
struct syscall {
	 unsigned short id;
	 char name[SYSCALL_NAME_LENGTH];
};
#define MAX_SYSCALLS 256
static struct syscall *syscall_table[MAX_SYSCALLS];
static int nsyscalls;

static struct namehash *hash;
static struct pidhash *pid_hash;
static struct replay_command* rcmd_pre = NULL; //Appended before actual command
static struct replay_command* rcmd_post = NULL; //Appended after actual command

#define PRINT_INDENT(...)
//#define PRINT_INDENT(...) ({int __j = indent_level; while (__j-- > 0) printf("	"); printf(__VA_ARGS__);})
void usage()
{
	fprintf(stderr, "Usage: trfs_pre [-c <key>] [-b <block-size>] [-d <key>] [-z -k <key file>] -f <trace-name>\n");
	fprintf(stderr, "	 -c <key> specifies that the file is checksummed\n");
	fprintf(stderr, "	 -b <block-size> specifies the block size of the checksum, default 4096\n");
	fprintf(stderr, "	 -z decompress using zlib\n");
	fprintf(stderr, "	 -d <key> specifies the decryption key\n");
	fprintf(stderr, "	 -k <key file> specifies the key file for decrypting the trace fields\n");
	fprintf(stderr, "	 -f specifies the file name\n");
	fprintf(stderr, "	 -m max combined memory buffers size\n");
	fprintf(stderr, "Note that order of -c and -z is important\n");
	exit(1);
}

void read_keys(char *keyfile)
{
	FILE *fp;
	char buf[256];
	struct decrypt_params keyp;
	int i;
	if ((fp = fopen(keyfile, "r")) == NULL) {
		fprintf(stderr, "Warning : could not open key file [%s] for reading\n", keyfile);
		return;
	}
	while (fgets(buf, 256, fp)) {
		char *name;
		char *key;
		buf[strlen(buf) - 1] = '\0';
		name = strtok(buf, ": ");
		key = strtok(NULL, ": ");
		if (name && key) {
			for (i = 0; i < ANON_MAX_FIELDS; i++) {
				if (strcmp(anon_keys[i].name, name) == 0) {
					if (make_key(&keyp, key)) {
						fprintf(stderr, "Warning: Invalid key\n");
						break;
					}
					anon_keys[i].valid = 1;
					memcpy(anon_keys[i].key, keyp.key, keyp.keylen);
					break;
				}
			}
			if (i == ANON_MAX_FIELDS) {
				fprintf(stderr, "Warning: Unknown field [%s]\n", name);
			}
		}
		else {
			fprintf(stderr, "Warning: Key file corrupt\n");
			fclose(fp);
			return;
		}
	}

	fclose(fp);
	return;


}

/*
 * This was created specifically for OP_READPAGES where a list of page indexes
 * and page buffers needs to be passed to replayfs.
 *
 * This function will append the page_list to replayfs by storing the list
 * of page indexes and the corresponding data buffers using append_data.
 * The data should look like:
 *	 ex. [PAGE_INDEX|PAGE_DATA_BUFFER|...]
 *	 PAGE_INDEX - the page index
 *	 PAGE_DATA_BUFFER - the data buffer responding to the page index
 *
 * TODO: Make the data buffers in the list reusable in the RAT
 */
int append_page_indexes(struct context* cont)
{
		char* buffer;
		int len = cont->args[1];
		int buf_len;
		int pos = 0;
		int count = 0;
		page_list_t* ptr = (void*)cont->args[2];
		page_list_t* prev;

		assert(len > 0);
		assert(ptr != NULL);

		buf_len = sizeof(long) * 2 * len;
		buffer = malloc(buf_len);
		if(buffer == NULL) {
			print_error_nomem();
			exit(ENOMEM);
		}

		//Write page indexes and the corresponding data buffers
		while(ptr != NULL) {
			memcpy(buffer+pos,&ptr->value,sizeof(long));
			pos += sizeof(long);
			memcpy(buffer+pos,&ptr->data_loc,sizeof(long));
			pos += sizeof(long);
			count++;

			prev = ptr;
			ptr = ptr->next;
			free(prev);
		}

		assert(count == len);
		cont->args[2] = res_manager_append_data(0, cont->time, buffer, buf_len,cont->pid);

		free(buffer);

		return 0;
}

int main(int argc, char *argv[])
{
	int err;
	struct parse_params params;
	int c;
	int index;
	memset(&params, 0, sizeof(params));
	params.cparams.blk_size = CKSUM_BLOCK_SIZE;

	prefp = fopen("/tmp/prefile","w");
	while ((c = getopt(argc, argv, "zc:b:f:d:k:m:")) != -1) {
		switch (c) {
		case 'c':
			if (params.n_ops == MAX_OPS) {
				fprintf(stderr, "Too many operations specified\n");
				usage();
			}
			params.ops[params.n_ops++] = OP_CKSUM_MD5;
			if (hmac_make_key(&params.cparams, optarg)) {
				fprintf(stderr, "Invalid key specified\n");
				usage();
			}
			break;
		case 'b':
			params.cparams.blk_size = atoi(optarg);
			break;
		case 'z':
			if (params.n_ops == MAX_OPS) {
				fprintf(stderr, "Too many operations specified\n");
				usage();
			}
			params.ops[params.n_ops++] = OP_UNCOMPRESS;
			break;
		case 'f':
			if (params.n_ops == MAX_OPS) {
				fprintf(stderr, "Too many operations specified\n");
				usage();
			}
			params.ops[params.n_ops++] = OP_FILE;
			params.file_name = optarg;
			break;
		case 'd':
			if (params.n_ops == MAX_OPS) {
				fprintf(stderr, "Too many operations specified\n");
				usage();
				}
			params.ops[params.n_ops++] = OP_DECRYPT;
			if (make_key(&params.dparams, optarg)) {
				fprintf(stderr, "Invalid key specified\n");
				usage();
			}
			break;
		case 'k':
			read_keys(optarg);
			break;
		case '?':
			fprintf(stderr, "Unknown option `-%c'.\n\n", optopt);
			usage();
			break;
		default:
			usage();
			break;
		}
	}
	index = optind;
	if (!params.file_name) {
		fprintf(stderr, "File name not specified\n");
		usage();
	}

	if (strlen(params.file_name) + 9 >= PATH_MAX) {
		fprintf(stderr, "File name is too long\n");
		exit(1);
	}

	if ((hash = create_namehash(1024)) == NULL) {
		fprintf(stderr, "Could not create name hash\n");
		exit(2);
	}

	if ((pid_hash = pidhash_create(1024)) == NULL) {
		fprintf(stderr, "Could not create PID hash\n");
		exit(3);
	}
	memset(syscall_table, 0, sizeof(syscall_table));

	res_manager_init(params.file_name, MAX_MEMORY_BUFFERS_SIZE);

	err = parse(&params);

	if (err != ERR_OK)
		fprintf(stderr, "Parsing failed with error code %d\n", err);

	res_manager_commit();

	pidhash_destroy(pid_hash);
	destroy_namehash(hash);
	fclose(prefp);
	return err;
}

struct descriptor {
	op_t type;
	char *str_type;
};



void *on_start_trace_message(op_t type, int pre_or_post, void* cont)
{
	struct context* context = cont;
	if(!pre_or_post) {
		res_manager_start_trace_message(type, cont);
		assert(context == NULL);
		context = malloc(sizeof(struct context));
		memset(context, 0, sizeof(*context));
		context->pid = PID_INVALID;
		context->ret = -1;
		context->status = 0;
	}
	PRINT_INDENT("BEGIN: %s %s\n", type_to_str_type(type), pre_or_post ? "POST_OP " : "PRE_OP");
	indent_level++;

	if (pre_or_post && (type == OP_LOOKUP || type == OP_LINK || type == OP_MKDIR)) {
		assert(context != NULL);
		struct namecomp *tmp = malloc(sizeof(struct namecomp));
		memset(tmp, 0, sizeof(*tmp));
		context->other = (void *)tmp;
	} else if (type == MSG_SYSCALL) {
		assert(context != NULL);
		struct syscall *tmp = malloc(sizeof(struct syscall));
		memset(tmp, 0, sizeof(*tmp));
		context->other = (void *)tmp;
	}
	return context;
}

int on_end_trace_message(op_t type, int pre_or_post, void *cont)
{
	struct context* context;
	assert(cont != NULL);
	context = cont;
	static int seq=0;
#if REPLAY_READS
	//Enable reads and disable readpage(s)
	if(type == OP_WRITE || type == OP_READPAGE || type == OP_READPAGES)
		goto exit;
#else
	/*
	 * OP_READ and OP_WRITE are redundant, READPAGE and COMMIT_WRITE are
	 * already handled.	At the time of this change, replayfs already skips
	 * OP_WRITE and OP_READ.
	 */
	if(type == OP_READ || type == OP_WRITE)
		goto exit;
#endif

	if(pre_or_post) {
		if(type == OP_D_RELEASE) {
			//add DPUT
			res_manager_add_dput_op(&rcmd_pre,context->time,context->saved[CONTEXT_SAVED_RAT_LOC],context->pid);
			assert(!res_manager_remove_unref_dentry(context->saved[CONTEXT_SAVED_RAT_LOC]));
			assert(!res_manager_add_unref_dentry(context->saved[CONTEXT_SAVED_RAT_LOC]));
		}
	}

	//if op == OP_READPAGES, then the page list must be added to the data segment
	if(type == OP_READPAGES)
		append_page_indexes(cont);

	if (context) {
		//For replayfs_writepage() and sync_page()
		res_manager_resolve_dentry_from_inum(context, type, pre_or_post, &rcmd_pre, &rcmd_post);
		//For reusing rat entries
		if (context->status == 0)
			res_manager_reuse_rat_entries(context, type, pre_or_post);

		if (pre_or_post && (type == OP_LOOKUP || type == OP_LINK || type == OP_MKDIR)) {
			struct namecomp *tmp = (struct namecomp *)context->other;
			if (context->other_fields_filled == NAMECOMP_NUMFIELDS)
				add_entry(hash, tmp->child_inum, tmp->parent_inum, tmp->name);
			if (tmp->name)
				free(tmp->name);
			free(tmp);
		} else if (type == MSG_SYSCALL && context->other_fields_filled == SYSCALL_NUMFIELDS) {
			syscall_table[nsyscalls++] = (struct syscall *)context->other;
		}
	}
	if (context->pid != PID_INVALID) {
		struct pidhash_entry *entry = pidhash_lookup(pid_hash, context->pid);
		if (context->pname)
			entry = pidhash_add_entry(pid_hash, context->pid, context->pname, context->len, entry);
		if (entry) {
			int i;
			PRINT_INDENT("Process Name: \"");
			for (i = 0; i < entry->len; i++) printf("%c", entry->pname[i]);
			printf("\"\n");
		} else
			PRINT_INDENT("PID = %d\n", context->pid);
	}
	indent_level--;
	PRINT_INDENT("END: %s %s\n", type_to_str_type(type), pre_or_post ? "POST_OP" : "PRE_OP");

	/*
	 * generate replay trace
	 */
	int i;
	struct replay_command rcmd;
	struct replay_command* rptr;

	memset(&rcmd, 0, sizeof(struct replay_command));

	//If OP is not below, it will not be replayed
	if (context->status == 0 && pre_or_post &&
			(type == OP_LOOKUP || type == OP_OPEN || type == OP_CREATE ||
			 type == OP_FLUSH || type == OP_RELEASE || type == OP_SYMLINK ||
			 type == OP_READPAGE || type == OP_COMMIT_WRITE || type == OP_LINK ||
			 type == OP_WRITEPAGE || type == OP_MKDIR || type == OP_RMDIR ||
			 type == OP_UNLINK || type == OP_D_RELEASE || type == OP_READDIR ||
			 type == OP_RENAME || type == OP_MKNOD || type == OP_SETATTR ||
			 type == OP_FSYNC || type == OP_SYNC_PAGE || type == OP_LLSEEK ||
			 type == OP_PERMISSION || type == OP_READ)) {
		fprintf(prefp,"\n# %d, OP: %d",++seq,type);
		rcmd.time = context->time;
		rcmd.command = type;
		rcmd.expected_retval = context->expected_retval;
		memcpy(rcmd.args, context->args, sizeof(context->args));
		rcmd.thread_loc = context->pid; //Thread_Manager will fix this later

#if MULTITHREADING
		thread_manager_request_thread(context->pid);
#endif

		/* root mount fillup where necessary */
		if (type == OP_OPEN || type == OP_CREATE || type == OP_PERMISSION) {
			rcmd.args[1] = RAT_ROOT_MNT;
		}
		rcmd.ret = context->ret;

		//if exist, add the pre replay_command
		if(rcmd_pre) {
			rcmd_pre->thread_loc = rcmd.thread_loc;
			//XXX: Assuming that all ops that uses rcmd_pre will return 0 on success
			if(rcmd.expected_retval == 0)
				res_manager_append_prog(rcmd_pre);
			else {
				print_warning_message("PRE_OP not added - expected return value:%d,op:%d",rcmd.expected_retval,rcmd.command);
				free(rcmd_pre);
			}
		}

		//Add the actual replay_command
		rptr = res_manager_append_prog(&rcmd);

		//if exist, add the post replay_command
		if(rcmd_post) {
			rcmd_post->thread_loc = rcmd.thread_loc;
			//XXX: Assuming that all ops that uses rcmd_post will return 0 on success
			if(rcmd.expected_retval == 0)
				res_manager_append_prog(rcmd_post);
			else {
				print_warning_message("POST_OP not added - expected return value:%d,op:%d",rcmd.expected_retval,rcmd.command);
				free(rcmd_post);
			}
		}

		for(i = 0;i < MAX_ARGS;i++)
			if(context->arg_str[i])
				res_manager_add_rat_ref(rptr,rptr->args[i]);
	}
	/*
	 * end of replay trace generation
	 */
/*	if(type == OP_SYNC_PAGE)
  {
    printf("DENTRY_LOC: %d, RAT_ROOT_MNT: %d, PAGE_INDEX:%d",rcmd.args[0],rcmd.args[1],rcmd.args[2]);
  }*/
 exit:
	if(rcmd_pre) {
		free(rcmd_pre);
		rcmd_pre = NULL;
	}
	if(rcmd_post) {
		free(rcmd_post);
		rcmd_post = NULL;
	}
	res_manager_end_trace_message(type,cont);
	if (context->pname)
		free(context->pname);
	if (pre_or_post) {
		free(context);
		context = NULL; //Mark context as freed
	}
	return 0;
}

void on_error(int error_type, int msg_num, char *desc)
{
	fprintf(stderr, "Error occurred while parsing trace: err_num = %d, msg_num = %d, description = %s\n", error_type, msg_num, desc);
}

void on_warning(int warn_type, char *desc)
{
	fprintf(stderr, "Warning: %s [%d]\n", desc, warn_type);
}

extern int get_decrypted_value(int anon_type, void *invalue, int inlen, void **outvalue, int *outl);
extern void release_decrypted_value(void *dvalue);


int on_data_field(op_t op, int pre_or_post, int tlv, field_t type, int length, void *value, void *cont)
{
	struct context* context = cont;
	void *dvalue = NULL;
	int err = ERR_OK;
	int anon_type = ANON_MAX_FIELDS;

	assert(cont != NULL);

	switch (type) {
		case ARG_UID:
			if (anon_keys[ANON_UID].valid) anon_type = ANON_UID;
			break;
		case ARG_GID:
			if (anon_keys[ANON_GID].valid) anon_type = ANON_GID;
			break;
		case ARG_PID:
			if (anon_keys[ANON_PID].valid) anon_type = ANON_PID;
			break;
		case ARG_PNAME:
			if (anon_keys[ANON_PNAME].valid) anon_type = ANON_PNAME;
			break;
		case ARG_PATHNAME:
			if (anon_keys[ANON_PATHNAME].valid) anon_type = ANON_PATHNAME;
			break;
	}
	if (anon_type != ANON_MAX_FIELDS) {
		err = get_decrypted_value(anon_type, value, length, &dvalue, &length);
		if (err != ERR_OK)
			goto out;
		value = dvalue;
	}

	switch (type) {
		case ARG_PAGE_INDEX:
			if (pre_or_post) {
				if (op == OP_READPAGE || op == OP_WRITEPAGE || op == OP_COMMIT_WRITE)
					context->args[2] = *(int*)value;
				else if(op == OP_SYNC_PAGE){
					context->args[2] = *(int*)value;
			        	}
			}
			break;

		case ARG_PAGES_INDEX:
			if(!pre_or_post && op == OP_READPAGES) {
				page_list_t* entry;
				page_list_t* ptr;

				entry = malloc(sizeof(page_list_t));
				memset(entry,0,sizeof(page_list_t));
				if(entry == NULL) {
					print_error_nomem();
					exit(ENOMEM);
				}
				entry->value = *(int*)value;
				entry->data_loc = -1;
				entry->next = NULL;
				ptr = (page_list_t*)context->args[2];

				if(ptr == NULL) {
					context->args[1] = 1;
					context->args[2] = (int)entry;
					break;
				} else {
					while(ptr->next != NULL)
					ptr = ptr->next;
					ptr->next = entry;
				}
				context->args[1]++;
			}
			break;

		case ARG_PATHNAME:
			if (pre_or_post) {
				/* zero terminate */
				char* name_buffer = malloc(length+1);
				if (!name_buffer) {
					fprintf(stderr, "Not enough memory\n");
					exit(1);
				}

				strncpy(name_buffer, value, length);
				name_buffer[length] = 0;

				if (op == OP_LOOKUP || op == OP_SYMLINK) {
					context->args[1] = res_manager_append_data(0, context->time, name_buffer, length + 1,context->pid);
					context->arg_str[1] = 1; //For memory buffer management
				}

				free(name_buffer);
			}
			break;

		case ARG_DATA:
			if (pre_or_post) {
				if (op == OP_READPAGE || op == OP_WRITEPAGE || op == OP_COMMIT_WRITE) {
					context->args[1] = res_manager_append_data(0, context->time, value, length,context->pid);
					context->arg_str[1] = 1;
				}
				else if(op == OP_READPAGES) {
					page_list_t* pptr = (page_list_t*)context->args[2];
					assert(context->args[2] != (int)NULL);

					while(pptr->next != NULL)
						pptr = pptr->next;
					assert(pptr->data_loc == -1); //ensures that entry is not already used
					pptr->data_loc = res_manager_append_data(0,context->time,value,length,context->pid);
				}
			}
			break;

		case ARG_INT_FLAGS:
			if (!pre_or_post) {
				if (op == OP_MKDIR || op == OP_CREATE || op == OP_READDIR || op == OP_MKNOD)
				  context->args[2] = *(int*)value;
				else if(op == OP_MKNOD)
			          context->args[3] = *(int*)value;
			}
			break;
			if (op == OP_READDIR || op == OP_FSYNC || op == OP_LLSEEK || op == OP_PERMISSION)
				context->args[3] = *(int*)value;
			break;

		case ARG_TIMESTAMP:
			{
				unsigned long secs;
				memcpy(&secs, value, sizeof(secs));
				context->time.secs = secs;
				//PRINT_INDENT("Time: %s\n", ctime(&secs));
			}
			break;

		case ARG_UTIMESTAMP:
			if (!pre_or_post) {
				unsigned long secs, usecs;
				memcpy(&secs, value, sizeof(secs));
				memcpy(&usecs, value + sizeof(secs), sizeof(usecs));
				//PRINT_INDENT("UTime: %ld, %ld\n", secs, usecs);
				context->time.secs = secs;
				context->time.usecs = usecs;
				assert(context->time.secs);
			}
			break;

		case ARG_INUM:
			context->saved[CONTEXT_SAVED_INUM] = *(int*)value;
			if(op == OP_LOOKUP || op == OP_SYNC_PAGE || op == OP_WRITEPAGE)
			        fprintf(prefp,"\nOP %d INUM: %d",op,*(int*)value);
			break;
		//This is added to capture the inode numbers of the children created to add to the
		//inode-dentry hash to be later back referenced during replay.
		//This will be called for OP_MKNOD. OP_MKDIR and OP_CREATE
		case ARG_CHILD_INUM:
			context->args[5] = *(int*)value;
            break;

		case ARG_ULONG:
			{
				if(!pre_or_post)
					break;
				if(op == OP_MKNOD) {
					/* XXX: tim: this is not correct, size is incorrect for some systems */
					context->args[2] = *(int*)value;
				} else if (op == OP_READ)
					context->args[2] = *(int*)value;
			}
			break;

		case ARG_PID:
			{
				pid_t pid;
				memcpy(&pid, value, sizeof(pid));
				context->pid = pid;
				//PRINT_INDENT("PID: %d\n", pid);
			}
			break;

		case ARG_IATTR:
			if(!pre_or_post)
				break;
			if(op == OP_SETATTR) {
				char* iattr_buf = malloc(length);
				if(!iattr_buf) {
					fprintf(stderr,"Not enough memory\n");
					exit(ENOMEM);
				}
				memcpy(iattr_buf,value,length);
				context->args[1] = res_manager_append_data(0,context->time,iattr_buf,length,context->pid);
				context->arg_str[1] = context->args[1]; //for memory buffer management

				free(iattr_buf);
			}
			break;

		case ARG_NEW_DENTRY_PTR:
		case ARG_DENTRY_PTR:
			context->saved[CONTEXT_SAVED_DENTRY_PTR] = *(int*)value;
			if (pre_or_post) {
				/*
				 * If op is lookup, ARG_DENTRY_PTR == the dentry to lookup
				 * and ARG_NEW_DENTRY_PTR == the parent dentry
				 */
				if (op == OP_LOOKUP && type == ARG_DENTRY_PTR) {
					context->saved[CONTEXT_SAVED_DENTRY_PTR] = *(int*)value;
					context->ret = res_manager_lookup_rat(*((int*)value), ARG_DENTRY);
					if (context->ret < 0)
					{
					  context->ret = res_manager_append_rat(*((int*)value), ARG_DENTRY);
					  if(context->ret != res_manager_lookup_rat(*((int*)value), ARG_DENTRY))
					     fprintf(prefp,"\nERROR1: Append RAT");

					  context->saved_fid_flags = context->saved_fid_flags | CONTEXT_SAVED_FID_DENTRY_GET;
					  fprintf(prefp,"\nLookup added %x to %d pos in RAT ",*((int*)value),context->ret);
					  //Add DGET
					  	res_manager_add_dget_op(&rcmd_post,context->time,context->ret,context->pid);
					}
					else
					{
					  fprintf(prefp,"\nLookup found %x at %d pos in RAT ",*((int*)value),context->ret);
					  if(res_manager_remove_unref_dentry(context->ret))
					  {
					    //Add DGET
					    res_manager_add_dget_op(&rcmd_post,context->time,context->ret,context->pid);
					  }
					}
					context->saved[CONTEXT_SAVED_RAT_LOC] = context->ret;
					//The dentry loc which has been added to rat needs to be stored in args[2].
					//This was missing and hence added it here.
					context->args[2] = context->ret;
				}
				if (op == OP_OPEN || op == OP_CREATE || op == OP_MKDIR ||
				op == OP_SYMLINK || op == OP_LINK || op == OP_RENAME ||
				op == OP_RMDIR || op == OP_UNLINK || op == OP_D_RELEASE ||
				op == OP_MKNOD || op == OP_SETATTR || op == OP_FSYNC ||
				op == OP_PERMISSION ||(type == ARG_NEW_DENTRY_PTR && op == OP_LOOKUP)) {
					int arg = 0;
					//This handles ops with two dentries
					if((op == OP_LINK && type == ARG_NEW_DENTRY_PTR) ||
					(op == OP_RENAME && type == ARG_NEW_DENTRY_PTR))
						arg = 1;
					context->saved[CONTEXT_SAVED_DENTRY_PTR] = *(int*)value;
					context->args[arg] = res_manager_lookup_rat(*((int*)value), ARG_DENTRY);
   				    if(op == OP_MKDIR || op == OP_CREATE)
						fprintf(prefp, "\nOP: %d dentry loc %d inode %d ",op,context->args[0],context->args[5]);

					if (context->args[arg] < 0)
					{
						/* We really dont know what to do here */
						fprintf(stderr,"Warning: dentry (0x%x) not found. Assuming \'/\'\n",*((int*)value));
						context->args[arg] = RAT_ROOT_DENTRY;
					}
					/*
					 * See if d_release has been called on this dentry if so
					 * call res_manager_add_dget_op() to add a dget op to
					 * it so we don't lose it beneath us.
					 */
					if(res_manager_remove_unref_dentry(context->args[arg])) {
						//Add DGET
						res_manager_add_dget_op(&rcmd_post,context->time,context->args[arg],context->pid);
					}
					context->saved[CONTEXT_SAVED_RAT_LOC] = context->args[arg];
				}
			}
			break;
		case ARG_FILE_PTR:
			context->saved[CONTEXT_SAVED_FILE_PTR] = *(int*)value;
			if (!pre_or_post) {
				if (op == OP_OPEN) {
					context->ret = res_manager_lookup_rat(*((int*)value), ARG_FILE);
					if (context->ret < 0)
						context->ret = res_manager_append_rat(*((int*)value), ARG_FILE);
					context->saved[CONTEXT_SAVED_RAT_LOC] = context->ret;
				}
				if (op == OP_FLUSH || op == OP_RELEASE || op == OP_READPAGE ||
				op == OP_WRITEPAGE || op == OP_COMMIT_WRITE || op == OP_READDIR ||
				op == OP_FSYNC || op == OP_LLSEEK || op == OP_READ) {
					int arg = 0;
					if(op == OP_FSYNC)
						arg = 1;
					context->args[arg] = res_manager_lookup_rat(*((int*)value), ARG_FILE);
					if (context->args[arg] < 0) {
						fprintf(stderr, "Warning: file (%p) not found. Skipping %s\n", (void*)(*(int*)value), type_to_str_type(op));
						context->status = -1;
					}
					context->saved[CONTEXT_SAVED_RAT_LOC] = context->args[arg];
				}
			}
			break;

		case ARG_FILE_FLAGS:
			if (!pre_or_post) {
				if (op == OP_OPEN) {
					context->args[2] = *(int*)value;
					//fprintf(stderr, "DEBUG: flags: 0x%x\n", get_rat_val(context->args[2]));
				} else if (op == OP_MKDIR) {
					context->args[1] = *(int*)value;
				}
			}
			break;

		case ARG_WRITE_FROM:
			if (pre_or_post) {
				if (op == OP_COMMIT_WRITE)
					context->args[3] = *(int*)value;
			}
			break;

		case ARG_WRITE_TO:
			if (pre_or_post) {
				if (op == OP_COMMIT_WRITE)
					context->args[4] = *(int*)value;
			}
			break;

			case ARG_LOFFSET:
			{
				if(!pre_or_post)
					break;
				if(op == OP_LLSEEK || op == OP_READ) {
					char* loff_buf = malloc(length);
					int arg = 1;
					if(op == OP_READ)
						arg = 3;
					if(!loff_buf) {
						fprintf(stderr,"Not enough memory\n");
						exit(ENOMEM);
					}
					memcpy(loff_buf,value,length);
					context->args[arg] = res_manager_append_data(0,context->time,loff_buf,length,context->pid);
					context->arg_str[1] = context->args[arg]; //For memory buffer managemen
					free(loff_buf);
				}
			}
			break;

		case ARG_RET:
			context->expected_retval = *((int*)value);
			break;
	}
out:
	if (dvalue)
		release_decrypted_value(dvalue);
	return err;
}

int on_header(struct stream_header *header)
{
	// do nothing
	return ERR_OK;
}

