/*
 * Copyright (c) 2003-2007 Erez Zadok
 * Copyright (c) 2003-2007 Stony Brook University
 * Copyright (c) 2003-2007 The Research Foundation of State University of New York
 * Copyright (c) 2004-2007 Nikolai Joukov
 * Copyright (c) 2003-2005 Akshat Aranya
 * Copyright (c) 2006-2007 Gopala Surya
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#include "gengraph.h"
#include "confparser.h"
#include "common.h"

/*
 * We divy out chunks of memory rather than call malloc each time so
 * we don't have to worry about leaking memory.  It's probably
 * not a big deal if all this memory was wasted but it this ever
 * goes into a library that would probably not be a good idea.
 */
#define NCHUNKS 16
#define CHUNK0SIZE 1024

#define strlcpy(x, y, z) \
        (strncpy((x), (y), (z)), \
         ((z) <= 0 ? 0 : ((x)[(z) - 1] = '\0')), \
         strlen((y)))


struct chunk {
	u_int n_left;
	void *m;
};

static struct chunk chunks[NCHUNKS];
static int cur_chunk;

static void *newchunk(u_int);
static void freechunks(void);
static inline struct block *new_block(int);

static void *
newchunk(u_int n)
{
	struct chunk *cp;
	int k, size;

#ifndef __NetBSD__
	/* XXX Round up to nearest long. */
	n = (n + sizeof(long) - 1) & ~(sizeof(long) - 1);
#else
	/* XXX Round up to structure boundary. */
	n = ALIGN(n);
#endif

	cp = &chunks[cur_chunk];
	if (n > cp->n_left) {
		++cp, k = ++cur_chunk;
		if (k >= NCHUNKS) {
			return NULL;
		}
		size = CHUNK0SIZE << k;
		cp->m = (void *)malloc(size);
		memset((char *)cp->m, 0, size);
		cp->n_left = size;
		if (n > size) {
			return NULL;
		}
	}
	cp->n_left -= n;
	return (void *)((char *)cp->m + cp->n_left);
}


static void
freechunks()
{
	int i;

	cur_chunk = 0;
	for (i = 0; i < NCHUNKS; ++i)
		if (chunks[i].m != NULL) {
			free(chunks[i].m);
			chunks[i].m = NULL;
		}
}

void tracefs_compile_done()
{
	freechunks();
}

/*
 * A strdup whose allocations are freed after code generation is over.
 */
char *
sdup(s)
        register const char *s;
{
        int n = strlen(s) + 1;
        char *cp = newchunk(n);

	strlcpy(cp, s, n);
        return (cp);
}


static inline struct block *
new_block(int type)
{
    struct block *p;

    p = (struct block *)newchunk(sizeof(*p));
    if (!p) {
	return NULL;
    }
    p->type = type;
    p->head = p;
    p->index = -1;

    return p;
}

/*
 * Backpatch the blocks in 'list' to 'target'.  The 'sense' field indicates
 * which of the jt and jf fields has been resolved and which is a pointer
 * back to another unresolved block (or nil).  At least one of the fields
 * in each block is already resolved.
 */
static void
backpatch(list, target)
    struct block *list, *target;
{
    struct block *next;

    while (list) {
        if (!list->sense) {
            next = JT(list);
            JT(list) = target;
        } else {
            next = JF(list);
            JF(list) = target;
        }
        list = next;
    }
}


/*
 * Merge the lists in b0 and b1, using the 'sense' field to indicate
 * which of jt and jf is the link.
 */
static void
merge(b0, b1)
	struct block *b0, *b1;
{
	register struct block **p = &b0;

	/* Find end of list. */
	while (*p)
		p = !((*p)->sense) ? &JT(*p) : &JF(*p);

	/* Concatenate the lists. */
	*p = b1;
}


void
gen_and(b0, b1)
	struct block *b0, *b1;
{
  debug("gen_and called\n");
	backpatch(b0, b1->head);
	b0->sense = !b0->sense;
	b1->sense = !b1->sense;
	merge(b1, b0);
	b1->sense = !b1->sense;
	b1->head = b0->head;
}

void
gen_or(b0, b1)
	struct block *b0, *b1;
{
  debug("gen_or called\n");
	b0->sense = !b0->sense;
	backpatch(b0, b1->head);
	b0->sense = !b0->sense;
	merge(b1, b0);
	b1->head = b0->head;
}

void
gen_not(b)
	struct block *b;
{
	b->sense = !b->sense;
}

struct block *
gen_retblk(int v)
{
    struct block *b = new_block(v ? TC_TRUE : TC_FALSE);

    b->u.ivalue = v;
    return b;
}

 struct block *
 gen_int_block(int t, int v)
{
  int type = 0;
  struct block *b;
  switch (t) {
  case CUID: type = TC_CUID; break;
  case CGID: type = TC_CGID; break;
  case FUID: type = TC_FUID; break;
  case FGID: type = TC_FGID; break;
  case SID:  type = TC_SID; break;
  case PID:  type = TC_PID; break;
  case PGRP:  type = TC_PGRP; break;
  case INO:  type = TC_INO; break;
  default: break; /* WILL NEVER HAPPEN */
  }
  debug("gen_int_block called with type = %s(%d), value = %d\n", get_node_type(type), type, v);
  b = new_block(type);
  b->u.ivalue = v;
  return b;
}


/*
 * generates a string block.
 * v should always be allocated using our own allocator (by using sdup)
 */

struct block *
gen_string_block(int t, char *v)
{
    int type = 0;
    struct block *b;
    switch (t) {
    case PNAME: type = TC_PNAME; break;
    case FILENAME: type = TC_NAME; break;
    case EXT: type = TC_EXT; break;
    default: break; /* WILL NEVER HAPPEN */
    }
    debug("gen_string_block called with type = %s(%d), value = %s [%d bytes]\n", get_node_type(type), type, v, strlen(v));
    b = new_block(type);
    b->u.svalue = v;
    return b;
}

int yyparse(void *);

struct graphset *tracefs_compile()
{
	struct graphset *gs = NULL;
	int err;

	if ((err = yyparse(&gs)) != 0) {
		fprintf(stderr, "yyparse: %d\n", err);
	}

	return gs;
}


struct user_output_driver *gen_file_destination(char *name, int buf, int sync_writes)
{
	struct user_output_driver *dest;

  	debug("gen_file_destination called: %s\n", name);

	dest = (struct user_output_driver *)newchunk(sizeof(struct user_output_driver));
	if (!dest) {
		errno = ENOMEM;
		return NULL;
	}

	dest->od_type = OD_FILE;
	dest->u.od_file.name = sdup(name);
	dest->u.od_file.buf_size = buf;
	dest->u.od_file.sync_writes = sync_writes;

	return dest;
}

struct user_output_driver *gen_async_destination(char *name, int buf, int count, int sync_writes)
{
	struct user_output_driver *dest;

  	debug("gen_async_destination called: %s\n", ((name) ? name : "NULL"));

	dest = (struct user_output_driver *)newchunk(sizeof(struct user_output_driver));
	if (!dest) {
		errno = ENOMEM;
		return NULL;
	}

	dest->od_type = OD_ASYNC;
	if (name)
		dest->u.od_async.name = sdup(name);
	else
		dest->u.od_async.name = NULL;
	dest->u.od_async.buf_size = buf;
	dest->u.od_async.buf_count = count;
	dest->u.od_async.sync_writes = sync_writes;

	return dest;
}
struct user_output_driver *gen_buffer_destination(int size)
{
	struct user_output_driver *dest;

  	debug("gen_buffer_destination called\n");

	dest = (struct user_output_driver *)newchunk(sizeof(struct user_output_driver));
	if (!dest) {
		errno = ENOMEM;
		return NULL;
	}

	dest->od_type = OD_BUFFER;
	dest->u.od_buffer.buf_size = size;
	return dest;
}

static char hex2byte(char *digits)
{
	char ret = 0;
	int i;
	for (i = 0; i < 2; i++) {
		ret = ret << 4;
		if ((digits[i] >= 'a' && digits[i] <= 'f')) {
			ret |= ((digits[i] - 'a' + 0xa) & 0xf);
		} else if ((digits[i] >= 'A' && digits[i] <= 'F')) {
			ret |= ((digits[i] - 'A' + 0xA) & 0xf);
		} else if ((digits[i] >= '0' && digits[i] <= '9')) {
			ret |= ((digits[i] - '0') & 0xf);
		}
	}
	return ret;
}

struct user_output_driver *gen_encrypt_destination(char *key, int ishexstr)
{
	struct user_output_driver *dest;
	int len;
  	debug("gen_encrypt_destination called\n");
	len = strlen(key);
	if (ishexstr && (len < 2 || len % 2 != 0)) {
		debug("Invalid hex string for key \n");
		errno = EINVAL;
		return NULL;
	}

	dest = (struct user_output_driver *)newchunk(sizeof(struct user_output_driver));
	if (!dest) {
		errno = ENOMEM;
		return NULL;
	}
	dest->od_type = OD_ENCRYPT;
	dest->u.od_encrypt.algorithm = ENCRYPT_BLOWFISH;

	if (ishexstr) {
		char *src;
		char *dst;
		dest->u.od_encrypt.keylen = (len - 2) / 2;
		dst = dest->u.od_encrypt.key = newchunk(dest->u.od_encrypt.keylen);
		src = key + 2;
		while (*src) {
			*dst = hex2byte(src);
			dst++;
			src += 2;
		}
	} else {
		dest->u.od_encrypt.keylen = len;
		dest->u.od_encrypt.key = sdup(key);
	}
	debug("key = %s, keylen = %d\n", dest->u.od_encrypt.key, dest->u.od_encrypt.keylen);
	return dest;
}


struct user_output_driver *gen_cksum_destination(char *key, int ishexstr)
{
	struct user_output_driver *dest;
	int len;
  	debug("gen_cksum_destination called\n");
	len = strlen(key);
	if (ishexstr && (len < 2 || len % 2 != 0)) {
		debug("Invalid hex string for key \n");
		errno = EINVAL;
		return NULL;
	}

	dest = (struct user_output_driver *)newchunk(sizeof(struct user_output_driver));
	if (!dest) {
		errno = ENOMEM;
		return NULL;
	}
	dest->od_type = OD_CKSUM;
	if (ishexstr) {
		char *src;
		char *dst;
		dest->u.od_cksum.keylen = (len - 2) / 2;
		dst = dest->u.od_cksum.key = newchunk(dest->u.od_cksum.keylen);
		src = key + 2;
		while (*src) {
			*dst = hex2byte(src);
			dst++;
			src += 2;
		}
	} else {
		dest->u.od_cksum.keylen = len;
		dest->u.od_cksum.key = sdup(key);
	}
	debug("key = %s, keylen = %d\n", dest->u.od_cksum.key, dest->u.od_cksum.keylen);

	return dest;

}
struct user_output_driver *gen_compress_destination(int buf_size)
{
	struct user_output_driver *dest;

  	debug("gen_compress_destination called\n");

	dest = (struct user_output_driver *)newchunk(sizeof(struct user_output_driver));
	if (!dest) {
		errno = ENOMEM;
		return NULL;
	}

	dest->od_type = OD_COMPRESS;
	dest->u.od_compress.in_buf_size = buf_size;

	return dest;

}

struct user_assembly_driver *gen_assembly(int type, int flags) {
	struct user_assembly_driver *dest;

  	debug("gen_assembly called: %d\n", type);

	dest = (struct user_assembly_driver *)newchunk(sizeof(struct user_assembly_driver));
	if (!dest) {
		errno = ENOMEM;
		return NULL;
	}

	switch(type) {
	case STREAM:
		dest->ad_type = AD_STREAM;
		dest->u.ad_stream.ad_verbosity = flags;
		break;
	case AGGREGATE:
		dest->ad_type = AD_AGGREGATE;
		break;
	default:
		fprintf(stderr, "Unknown assembly driver!\n");
	}

	return dest;

}

int add_destination(struct graphset **graphset, struct block *root, struct array *adlist, struct array *odlist)
{
	struct graphset *ogs = *((struct graphset **)(graphset));
	struct graphset *gs;

	gs = newchunk(sizeof(struct graphset));
	if (!(gs)) {
		errno = ENOMEM;
		return -1;
	}
	memset(gs, 0, sizeof(*gs));

	if (root)
		gs->root = finish_parse(root);
	gs->ods = (struct user_output_driver *)odlist->data;
	gs->n_ods = odlist->count;
	gs->ads = (struct user_assembly_driver *)adlist->data;
	gs->n_ads = adlist->count;
	gs->next = ogs;

	*((struct graphset **)(graphset)) = gs;

	return 0;
}

struct block *
finish_parse(p)
	struct block *p;
{
	backpatch(p, gen_retblk(1));
	p->sense = !p->sense;
	backpatch(p, gen_retblk(0));
	return p->head;
}

struct array *make_array(void *item, int elsize) {
	struct array *ar = newchunk(sizeof(struct array));

	ar->count = 1;
	ar->data = newchunk(elsize);
	memcpy(ar->data, item, elsize);


	return ar;
}

struct array *append_array(struct array *ar, void *item, int elsize) {
	void *olddata;

	olddata = ar->data;
	ar->data = realloc(olddata, (ar->count + 1) * elsize);
	memcpy(ar->data + elsize * ar->count, item, elsize);
	ar->count++;

	return ar;
}

char *gen_bmap(int bit)
{
	char *bmap = (char *)newchunk(BITMAP_SIZE);
	memset(bmap, 0, sizeof(bmap));
	OP_SET_ENABLED(bmap, bit);
	return bmap;
}

char *set_bit(char *bmap, int bit)
{
	OP_SET_ENABLED(bmap, bit);
	return bmap;
}

struct block *gen_bmap_block(int t, char *v)
{
    struct block *b;
    debug("gen_bmap_block called\n");
    b = new_block(t);
    b->u.svalue = v;
    return b;
}
